import java.io.*;
import java.net.*;

public class Network
{
	private String host;
	private int port;
	private Socket sock;
	
	public static final String proto[] = { "telnet", "ssh" };
	public static final int TELNET = 0;
	public static final int SSH = 1;
	
	private int protocol;
	private boolean authenticated;
	private boolean connected, closed;
	
	private ch.ethz.ssh2.Connection conn;
	private ch.ethz.ssh2.Session sess;
	
	private InputStream is;
	private OutputStream os;
	
	public Network( String h, int p, int pr )
	{
		host = h;
		port = p;
		protocol = pr;
		
		if( protocol == SSH ) {
			connected = false;
			authenticated = false;
			closed = false;
		}
	}

	public boolean connect()
	{
		// System.out.println("Network.connect()");
		if( protocol == TELNET ) {
			try {
				System.out.println("Connecting to host: " + host + ", port: " + port + " ...");
				sock = new Socket( host, port );
				
				// disable Nagle's algorithm
				// 不要把很多小封包包裝成一個大封包再一次送出，
				// 這非常重要，連續顯示的游標是使用者評斷顯示速度的重要依據。
				sock.setTcpNoDelay( true );
				
			} catch ( UnknownHostException e ) {
				// 可能是未連線或連線位置錯誤
				// e.printStackTrace();
				System.out.println("Caught UnknownHostException in Network::connect()");
				return false;
			} catch ( IOException e ) {
				// e.printStackTrace();
				System.out.println("Caught IOException in Network::connect()");
				return false;
			}

			try {
				is = sock.getInputStream();
				os = sock.getOutputStream();
			} catch ( IOException e ) {
				e.printStackTrace();
				return false;
			} 
			
			return true;

		} else if( protocol == SSH ) {
			try {
				System.out.println("SSH2 to host: " + host + ", port: " + port + " ..." );
				conn = new ch.ethz.ssh2.Connection( host, port );
				// FIXME: magic number, 應可自行設定 timeout
				conn.connect( null, 5000, 60000 );
				conn.setTCPNoDelay( true );
				connected = true;
				
				String username = Auth.getUsername();
				if( username == null ) {
					disconnect();
					return false;
				}
				
				String[] methods = conn.getRemainingAuthMethods( username );
				System.out.println("Available ssh2 authentication methods:");
				if( methods != null ) {
					for(int i = 0; i < methods.length; i++) {
						System.out.println( methods[i] );
					}
				} else {
					System.out.println( "getRemainingAuthMethods( username ) = null" );
				}
				
				if( conn.isAuthMethodAvailable( username, "password" ) ) {
					authenticated = conn.authenticateWithPassword( username, Auth.getPasword( "輸入密碼", "請輸入密碼：" ) );
					System.out.println("SSH2: password auth");
				} else if( conn.isAuthMethodAvailable( username, "keyboard-interactive" )) {
					System.out.println("SSH2: keyboard-interactive auth");
					authenticated = conn.authenticateWithKeyboardInteractive( username, new Auth() );
				} else if( conn.isAuthMethodAvailable( username, "publickey" ) ) {
					System.out.println("SSH2 publickey(not yet support)");
				} else {
					System.out.println("unknowd SSH2 authentication method.");
				}
				
				if( authenticated == false ) {
					System.out.println("authentication failed.");
					disconnect();
					return false;
				}
				
				sess = conn.openSession();
				// FIXME: magic number
				sess.requestPTY( "vt100", 80, 24, 0, 0, null );
				// sess.requestPTY( "xterm", 80, 24, 0, 0, null );
				
				sess.startShell();
				is = sess.getStdout();
				os = sess.getStdin();

				return true;
				
			} catch (IOException e) {
				e.printStackTrace();
				connected = false;
				disconnect();
				return false;
			}
		} else {
			// Unknown protocol
			return false;
		}

	}

	public void disconnect()
	{
		if( protocol == TELNET ) {
			// 如果跟本沒連線成功或是連線已被關閉則不做任何事。
			if( sock == null || sock.isClosed() ) {
				return;
			}
			
			try {
				is.close();
				os.close();
				sock.close();				
			} catch( IOException e ) {
				System.out.println( "Catch IOException when disconnect!");
				e.printStackTrace();
			}
			
			if( sock.isClosed() ) {
				// System.out.println( "Connection closed!");
			} else {
				System.out.println( "Disconnect failed!");
			}
			
		} else if( protocol == SSH ) {
			
			// 不論是否連線成功都要 close connection
			conn.close();
			
			// 身份認證成功則需要 close
			if( authenticated ) {
				try {
					is.close();
					os.close();
					sess.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
			
			closed = true;

		}
	}
	
	public void writeByte( byte b ) throws IOException
	{
		os.write( b );
		os.flush();
	}
	
	public void writeBytes ( byte[] buf, int offset, int size ) throws IOException
	{
		os.write( buf, offset, size );
		os.flush();
	}
	
	/*
	public void write( char c )
	{
		try {
			osw.write( c );
			osw.flush();
		} catch( IOException e ) {
			disconnect();
			// e.printStackTrace();
		}
	}

	public void writeChars( char[] buf, int offset, int size )
	{
		try {
			osw.write( buf, offset, size );
			osw.flush();
		} catch( IOException e ) {
			disconnect();
			// e.printStackTrace();
		}
	}
	*/
	
	public int readByte() throws IOException
	{
		int r;

		r = is.read();
		if( r == -1 ) {
			// System.out.println("read -1 (EOF), disconnect().");
			throw new IOException();
		}
		return r;
	}
	
	public int readBytes( byte[] b ) throws IOException
	{
		int r;

		r =  is.read( b );
		
		if( r == -1 ) {
			// System.out.println("read -1 (EOF), disconnect().");
			throw new IOException();
		}
		
		return r;
	}

	public boolean isClosed() {
		
		if( protocol == TELNET ) {
			// XXX: 連線建立失敗也當成已關閉連線。
			if( sock == null ) {
				return true;
			}

			return sock.isClosed();
		} else {
			return closed;
		}
	}

	public boolean isConnected() {
		if( protocol == TELNET ) {
			if( sock == null ) {
				return false;
			}
			return sock.isConnected();
		} else {
			return connected;
		}
	}

}
